/**
 * @name:    Game of Life
 *
 * @brief    Game of Life Application
 *
 * @author: Tomasz Gebarowski <gebarowski@gmail.com>
 * @date:   Wednesday 03 Nov 2008, 8:59 AM
 */
#include <gol_common.h>
#include <gol_board.h>
#include <gol_sub_board.h>
#include <pp_timer.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <mpi.h>
#include <unistd.h>

/**
 * @brief Initial population in percent
 *        which is number of initially populated cells
 */
#define INITIAL_POPULATION 30


/**
 * @brief Display final output ??
 */
//#define DISPLAY

/**
 * @brief  Board X and Y size
 */
#define GOL_BOARD_SIZE    10

/**
 * @brief   Number of population steps
 */
#define POPULATION_STEPS  2

/**
 * @brief    Random Seed 
 */
#define SEED              111

/** -------------------------------------------------------------------------*/
/** ------------------------------- Typedefs  -------------------------------*/
/** -------------------------------------------------------------------------*/

typedef enum
{
    TAG_FIRST_ROW = 0,
    TAG_LAST_ROW  = 1,
    TAG_FINAL_RESULT = 2
} row_tag_t;


/**
 * @param Parse command line parameters
 */
bool static 
__parse_cmd_line_args(int argc,
                      char *argv[],
                      unsigned int *size,
                      unsigned int *seed,
                      unsigned int *steps)
{
    bool ret_val = TRUE;

    if (argc == 4)
    {
        unsigned int tmp_size  = atoi(argv[1]);
        unsigned int tmp_seed  = atoi(argv[2]);
        unsigned int tmp_steps = atoi(argv[3]);

        if (tmp_size < 2 || tmp_size > 1000)
        {
            gol_message("Game Board size must be inside interval <2;1000>\n");
            ret_val = FALSE;
        }

        if (tmp_seed < 1 && tmp_seed > 10000)
        {
            gol_message("Initial population must be inside <1;10000>\n");
            ret_val = FALSE;
        }

        if (tmp_steps < 1 && tmp_steps > 10000)
        {
            gol_message("Population steps must be inside <1;10000>\n");
            ret_val = FALSE;
        }

        if (ret_val)
        {
            if (size != NULL)
                *size = tmp_size;

            if (seed != NULL)
                *seed = tmp_seed;

            if (steps != NULL)
                *steps = tmp_steps;
        }
    }
    else
    {
        gol_message("Usage %s [GAME_BOARD_SIZE] [SEED] [STEPS]\n", argv[0]);
        ret_val = FALSE;
    }

    return ret_val;
}

/** -------------------------------------------------------------------------*/
/** ----------------------------------- Main --------------------------------*/
/** -------------------------------------------------------------------------*/

int main(int argc, char *argv[])
{
    int rank, size;

    /** Game Board Width */
    unsigned int game_width  = GOL_BOARD_SIZE;
    unsigned int game_height = GOL_BOARD_SIZE;

    /** Execution timer */
    pp_timer_t *p_timer = pp_timer_create();

    /** Current Sub Board instance */
    gol_sub_board_t *p_sub_board =  NULL;

    /** Sub Board Height */
    unsigned int sub_height = 0;

    /** Initial population in percent */
    unsigned init_pop = INITIAL_POPULATION;

    /** Number of populations steps */
    unsigned int population_steps = 0;
    
    /* Static seed for randomization */
    unsigned int seed = 0;

    /** Average Number of rows of Game Board per process */
    unsigned int n_rows_per_proc = 0;

    MPI_Init( &argc, &argv );
    MPI_Comm_rank( MPI_COMM_WORLD, &rank );
    MPI_Comm_size( MPI_COMM_WORLD, &size );

    /** Parse Command Line parameters only in Root Process */
    if (rank == 0)
    {
        if ( !__parse_cmd_line_args(argc,
                                    argv,
                                    &game_width,
                                    &seed,
                                    &population_steps))
        {
            MPI_Finalize();
            return 66;
        }
    }
   
    /** Broadcast parameters to all processes */
    MPI_Bcast(&game_width, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&seed,       1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&population_steps, 1, MPI_INT, 0, MPI_COMM_WORLD);

    game_height = game_width;

    /** Calculate average number of rows per process*/
    n_rows_per_proc = game_height / size;

    if ( size > game_height )
    {
        gol_message("Number of processes cannot be greater then rows number\n");
        MPI_Finalize();
        return - 1;
    }

    sub_height = n_rows_per_proc;

    /** For the last process calculate remaining number of rows */
    if (rank == size - 1  &&
        size > 1)
    {
        sub_height = game_height - (size - 1) * n_rows_per_proc;
    }

    /** Create Sub Board */
    p_sub_board = gol_sub_board_create(sub_height, 
                                       game_width);

    if (p_sub_board != NULL)
    {
        gol_board_t  *p_gol_board = NULL;

        MPI_Request request_arr[2];
        MPI_Status  status_arr[2];

        unsigned int first_row[game_width];
        unsigned int last_row[game_width];
        unsigned int step;
        unsigned int alive_cnt     = 0;
        unsigned int sub_alive_cnt = 0;

        if (rank == 0)
        {
            gol_message("Game Board [%d x %d] Seed: %d Steps: %d\n",
                         game_width, game_height, seed,population_steps);
            gol_message("Processing...\n");
            p_gol_board = gol_board_create(game_height,game_width);
        }

        gol_sub_board_init(p_sub_board,
                           seed,
                           n_rows_per_proc * rank * game_width,
                           init_pop);

        /** Start timer for Root Thread */
        if (rank == 0)
        {
            if (p_timer != NULL)
                pp_timer_start(p_timer);
        }



        /** Perform action POPULATION_STEPS times */
        for (step = 0; step < population_steps; step++)
        {
            /** ---------- Exchange Neighbours ----------- */
            MPI_Irecv( last_row, 
                       game_width,
                       MPI_INT,
                       gol_prev_rank(rank,size),
                       TAG_LAST_ROW,
                       MPI_COMM_WORLD,
                       &request_arr[0] );

            MPI_Irecv( first_row, 
                       game_width,
                       MPI_INT,
                       gol_next_rank(rank,size),
                       TAG_FIRST_ROW,
                       MPI_COMM_WORLD,
                       &request_arr[1] );

            /** Send first row to rank-1 processs */
            MPI_Send( p_sub_board->p_arr, 
                      game_width,
                      MPI_INT,
                      gol_prev_rank(rank,size),
                      TAG_FIRST_ROW,
                      MPI_COMM_WORLD );

            /** Send last row to  rank+1 process  */
            MPI_Send( gol_last_row_ptr(p_sub_board->p_arr, sub_height, game_width ),
                      game_width,
                      MPI_INT,
                      gol_next_rank(rank,size),
                      TAG_LAST_ROW,
                      MPI_COMM_WORLD );

            /** Wait for both last and first raw neighbours */
            MPI_Waitall(2, request_arr, status_arr );
    
            /** Create temporary sub-game boards */
            gol_sub_board_create_tmp_src(p_sub_board, first_row, last_row, game_width );
       
            /** Run single iteration step*/
            gol_sub_board_run_step(p_sub_board);
        }

        /** Compute Number of Alive cells in sub boards */
        gol_sub_board_count_alive(p_sub_board,&sub_alive_cnt);

        /** Sum alive cells from sub boards */
        MPI_Reduce(&sub_alive_cnt, &alive_cnt, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);

        if (rank == 0)
        {
            if (p_timer != NULL)
                pp_timer_stop(p_timer);
        }

        #ifdef DISPLAY
        /** Gather results */
        MPI_Gather(p_sub_board->p_arr, p_sub_board->rows * p_sub_board->cols, MPI_INT,
                   (p_gol_board != NULL) ? p_gol_board->p_arr : NULL,
                    p_sub_board->rows * p_sub_board->cols, MPI_INT, 0, MPI_COMM_WORLD);
        #endif

        if (rank == 0)
        {
            if (p_timer != NULL)
            {
                gol_message("Measured time: %d msecs\n", pp_timer_measured_time(p_timer));
            }
            gol_message("Total number of alive cells:%d\n", alive_cnt);

        #ifdef DISPLAY
            gol_board_print(p_gol_board);
        #endif
            gol_board_dispose(p_gol_board);
        }


        /** Dispose Sub Game Board */
        gol_sub_board_dispose(p_sub_board);
    }
    else
    {
        gol_message("Cannot create gol_sub_board_t object");
    }               

    MPI_Finalize();
    return 0;
}
