/**
 * @name:    Game of Life
 *
 * @brief    Game of Life sub_board implementation
 *
 * @author: Tomasz Gebarowski <gebarowski@gmail.com>
 * @date:   Wednesday 03 Nov 2008, 8:59 AM
 */
#include <gol_common.h>
#include <gol_sub_board.h>
#include <limits.h>
#include <string.h>




/** -------------------------------------------------------------------------*/
/** -------------------------- Private Typededs -----------------------------*/
/** -------------------------------------------------------------------------*/

/**
 * @brief Defines realtive position of a neighbour 
 *        with respect to investigated cell
 */
typedef struct
{
    short dx;
    short dy;
} gol_neighbour_t;

/** -------------------------------------------------------------------------*/
/** -------------------------- Predefinitions  ------------------------------*/
/** -------------------------------------------------------------------------*/


static gol_status_t
__gol_board_process_cell(gol_sub_board_t *self,
                         unsigned int i,  
                         unsigned int j,    
                         gol_state_t *cell);

/** -------------------------------------------------------------------------*/
/** -------------------------- Global Variables -----------------------------*/
/** -------------------------------------------------------------------------*/


/**
 * @brief Definition of all possible neighbors of a cell
 */
static gol_neighbour_t g_nbs[8] = { {-1,-1}, {0,-1}, {1,-1},
                                  {-1,0} ,         {1,0},
                                  {-1,1} , {0,1},  {1,1}
                                  };

/** -------------------------------------------------------------------------*/
/** ------------------------ Function Implementation ------------------------*/
/** -------------------------------------------------------------------------*/

/**
 * @brief Allocates memory for Game Board
 */
gol_sub_board_t *
gol_sub_board_create(unsigned int rows,
                     unsigned int cols)
{
    gol_sub_board_t * p_sub_board = malloc(sizeof(gol_sub_board_t));

    if (p_sub_board != NULL)
    {
        p_sub_board->rows = rows;
        p_sub_board->cols = cols;
        p_sub_board->p_arr = calloc(rows * cols, sizeof(unsigned int));

        if (p_sub_board->p_arr == NULL)
        {
            gol_message("Cannot allocate memory for Game Board");
            free(p_sub_board);
            return NULL;
        }
    }
    return p_sub_board;
}

/**
 * @brief Initializes Game Board
 */
gol_status_t
gol_sub_board_init(gol_sub_board_t *self,
                   unsigned int seed,
                   unsigned int offset,
                   unsigned int init_popul)
{
    if (self != NULL)
    {
        unsigned int i,j;
        float alive_prob = init_popul / 100.0f;

        srandom(seed);

        self->p_tmp_src = NULL;

        /** Skip a given number of random values */
        for(i = 0; i < offset; i++)
        {
            random();
        }

        for(i = 0; i < self->rows; i++)
        {
            for(j = 0; j < self->cols; j++)
            {
                gol_state_t state = GOL_DEAD;
                float rand_val =  (1.0 * random() / RAND_MAX);

                if (rand_val < alive_prob) 
                {  
                    state = GOL_ALIVE;
                }

                self->p_arr[self->cols *i + j] = state;
            }
        }
    }
    else
    {
        return GOL_EINVPTR;
    }

    return GOL_OK;
}

/**
 * @brief Creates Source Sub Board by concatenating
 *        first_row with sub_board and p_last_row
 */
gol_status_t
gol_sub_board_create_tmp_src(gol_sub_board_t *self, 
                             unsigned int *p_first_row,
                             unsigned int *p_last_row,
                             unsigned int  row_size)
{
    gol_status_t ret_code = GOL_OK;

    if  (self != NULL &&
         p_first_row != NULL &&
         p_last_row  != NULL)
    {
        if (self->cols == row_size)
        {
            unsigned int arr_size = self->rows * self->cols;

            if (self->p_tmp_src != NULL)
                free(self->p_tmp_src);

            self->p_tmp_src = malloc(sizeof(unsigned int) * (row_size * 2 + arr_size));

            if (self->p_tmp_src != NULL)
            {
                /** Append first row */
                memcpy(self->p_tmp_src, p_last_row, row_size * sizeof(unsigned int));
                /** Append current cells */
                memcpy(&(self->p_tmp_src[row_size]), self->p_arr, arr_size * sizeof(unsigned int));
                /** Append second row */
                memcpy(&(self->p_tmp_src[arr_size + row_size]), p_first_row, row_size * sizeof(unsigned int));

            }
            else
            {
                gol_message("Not enough memory\n");
                ret_code = GOL_ENOMEM;
            }
        }
        else
        {
            gol_message("Wrong size of a row\n");
            ret_code = GOL_ERROR;
        }
    }
    else
    {
        gol_message("Invalid pointers self(%p) p_first_row(%p) p_last_row(%p)\n",
                     self, p_first_row, p_last_row);
        ret_code = GOL_EINVPTR;
    }

    return ret_code;
}


/**
 * @brief Runs a single iteration of game logic
 */
gol_status_t
gol_sub_board_run_step(gol_sub_board_t *self)
{
    if (self != NULL)
    {
        if (self->p_arr != NULL &&
            self->p_tmp_src)
        {
            int i,j;
            for( i = 0; i < self->rows; i++)
            {
                for( j = 0; j < self->cols; j++)
                {
                    gol_state_t  cell_state = 0;
                    gol_status_t ret_code = __gol_board_process_cell(self,  /** Pointer to Board */
                                                                     i,     /** X pos of cell */
                                                                     j,     /** Y pos of cell */
                                                                     &cell_state); /** Cell Output state*/
            
                    if (ret_code != GOL_OK)
                    {
                        gol_sub_board_dispose(self);
                        return ret_code;
                    } 
                    /** Change state of cell in dst game board*/
                    else
                    {
                        self->p_arr[self->cols * i + j] = cell_state;
                    }
                }
            }
        }
    }
    else
    {
        return GOL_EINVPTR;
    }
    return GOL_OK;
}


/**
 * @brief Count total alive cellshttp://www.onet.pl/
 */
gol_status_t
gol_sub_board_count_alive(gol_sub_board_t *self,
                          unsigned int    *p_total_alive)
{
    unsigned int counter = 0;

    if (self != NULL &&
        p_total_alive != NULL)
    {
        unsigned int i,j;

        if (self->p_arr != NULL)
        {
            for( i = 0; i < self->rows; i++)
            {
                for( j = 0; j < self->cols; j++)
                {
                    if ( self->p_arr[self->cols * i + j] == GOL_ALIVE)
                    {
                        counter++;
                    }
                }
            }
            *p_total_alive = counter;
            return GOL_OK;
        }
    }
    return GOL_EINVPTR;
}


/**
 * @brief Disposes object 
 */
void
gol_sub_board_dispose(gol_sub_board_t *self)
{
    if (self != NULL)
    {
        if (self->p_arr != NULL)
        {
            free(self->p_arr);
        }

        if (self->p_tmp_src != NULL)
        {
            free(self->p_tmp_src);
        }

        free(self);
    }
}

/** ----------------------------- Private Functions --------------------------*/

/**
 * @brief Performs a processing iteraton on a single cell
 *
 * @param self Pointer to instance of Game Board
 * @param i    X-coordinate of cell
 * @param j    Y-coordinate of cell
 * @param cell (OUT) State of changed cell
 */

static gol_status_t
__gol_board_process_cell(gol_sub_board_t *self,
                         unsigned int i,  
                         unsigned int j,    
                         gol_state_t *cell)
{

    if (self != NULL &&
        cell != NULL)
    {
        if (self->p_arr != NULL)
        {
            /** Counter of Alive Neighbours */
            unsigned int n_alive_nbs = 0;
            unsigned int n = 0;
            
            /** Foreach Neighbor */
            for(n = 0; n < sizeof(g_nbs) / sizeof(gol_neighbour_t); n++)
            {               
                /** Get neighbor cell */
                unsigned int r = i + 1 + g_nbs[n].dx;
                unsigned int c = gol_mod( j + g_nbs[n].dy, self->cols);

                /** If neighbor is alive then increment counter */
                if (  self->p_tmp_src[(self->cols) * r + c] == GOL_ALIVE )
                {
                    n_alive_nbs++;
                }
            }

            /** Decision if the cell should be killed, revived or remain in the same state*/

            /** Remain in the same state */
            if (n_alive_nbs == 2)
            {
                *cell =  self->p_tmp_src[(self->cols) * i + j];
            }
            /** Revive Cell */
            else if (n_alive_nbs == 3)
            {
                *cell = GOL_ALIVE;
            }
            /** Kill Cell */
            else
            {
                *cell = GOL_DEAD;
            }
            return GOL_OK;
        }
    }
    else
    {
        gol_message("Invalid ptr self(%p) cell(%p)\n", 
                     self,
                     cell);
    }

    return GOL_EINVPTR;
}




